"""
Folder that contains connectors to different libraries:
- bottleneck: bottleneck library for fast computation of moving statistics
- babachi: babachi library for signal segmentation
- bam2bed: bam2bed library for conversion of bam files to bed files
"""